% Copyright  2006 Gregory Berchin, all rights reserved.  Copying permitted
% for personal, non-commercial use provided that this copyright statement
% remains intact.  Commercial use prohibited without expressed written
% consent from Gregory Berchin.

% Revision record
%   04/28/2006  GB  original; version 1.00
%   04/29/2006  GB  added fclose(); version 1.01
%   05/15/2006  GB  changed some variable names for agreement with magazine
%                   article, added explanation of weighting; version 1.02
%   06/14/2006  GB  more variable name changes for agreement with magazine
%                   article; version 1.03
%   12/16/2006  GB  added plots, requires Signal Processing Toolbox;
%                   version 1.04
%	12/21/2006	GB	revised implementation of artificial delay "delta" so 
%					that delay is excluded from "u" sequence -- necessary 
%					to avoid unstable pole placement at PI radians that 
%					almost always results when u(k) contains any sine 
%					components; version 2.00


% Input file format:
%
%   M   samplerate
%   frequency(1)    magnitude(1)    phase(1)    weight(1)
%   frequency(2)    magnitude(2)    phase(2)    weight(2)
%   ...
%   frequency(M)    magnitude(M)    phase(M)    weight(M)
%
% M is the number of measurements; 
% samplerate and frequency are in Hz; 
% phase is in degrees.


% This program implements WEIGHTED least-squares.  Standard weight is 1.0.
% A weight of 2.0 means that the associated sample is mathematically 
% treated as though it appears twice, and so on.  Thus, higher weighting 
% of a particular sample means that it is given greater importance in the 
% least-squares fit.


% Normally a transfer function would look like this:
%                               -1            -N
%                        b + b z   + ... + b z
%        -delta   Y(z)    0   1             N
%       z       * ---- = -----------------------
%                 U(z)           -1           -D
%                         1 + a z  + ... + a z
%                              1            D
%
% But this is Matlab, with its weird indexing, so this is what has to be used:
%                               -1              -N
%                        b + b z   + ... + b   z
%        -delta   Y(z)    1   2             N+1
%       z       * ---- = -------------------------
%                 U(z)           -1             -D
%                         1 + a z  + ... + a   z
%                              2            D+1
%
% A quantity of time delay equal to the "delta" term can be artificially 
% added to the frequency response data to improve algorithm convergence.  
% The delay will manifest itself in the transfer function as near-zero 
% values for the first "delta" numerator (b) coefficients.  They will not 
% be exactly zero because of the least-squares nature of the fit.


close all;


% Open input file
infile = input('Enter name of input file:  ','s');
infid = fopen(infile,'r');

% Input number of measurement points "M" and sampling rate
temp = fscanf(infid,'%d %f',2);   % [2 x 1]
M = temp(1)    % scalar
samplerate = temp(2)    % scalar
ts = 1.0/samplerate; % sample period; scalar
halfrate = samplerate / 2.0;    % scalar
clear temp;

% Input numerator order "N"
N = input('Enter numerator order "N":  ')    % scalar
if N < 0
    error('Negative numerator order.');
end

% Input denominator order "D"
D = input('Enter denominator order "D":  ')    % scalar
if D < 0
    error('Negative denominator order.');
end

% System should be overspecified
if M < (D+N+1)
    error('System is underspecified (too few equations in too many unknowns).')
end

% Input artifical delay
delta = input('Enter artificial delay (number of samples):  ')    % scalar
if delta < 0
    error('Negative delay.');
end

% User reassurance
disp(' ');
disp('  Working ...');

% Initialize some constants
format long g;
DTR = pi / 180.0;    % scalar
TWOPI = 2.0 * pi;    % scalar

% Initialize some matrix dimensions
u = zeros(1,N+1);
X = zeros(M,D+N+1);
if (D ~= 0)
    y = zeros(1,D);
end

% Input frequency, magnitude, phase, weight vectors
temp = fscanf(infid,'%f %f %f %f',[4,M]);   % [4 x M]; fscanf fills column-wise but input data are stored row-wise
fclose(infid);
frequency = temp(1,:)';  % transpose row 1; [M x 1]
magnitude = temp(2,:)';  % transpose row 2; [M x 1]
phase = temp(3,:)';  % transpose row 3; [M x 1]
weight = temp(4,:)';  % transpose row 4; [M x 1]
clear temp;

% Convert phase from degrees to radians
inputphase = phase;   % [M x 1]
phase = phase .* DTR;   % [M x 1]

% Add artificial delay to phase
phase = phase - ((delta * TWOPI / samplerate) .* frequency);   % [M x 1]

% Create X matrix
for m = 1:M
    if frequency(m) > halfrate
        % While mathematically valid, this probably isn't what was intended.
        error('Frequency is greater than half the sampling rate.');
	end
	omega_ts = TWOPI * frequency(m) * ts;  % scalar
    if (D ~= 0) % denominator order not equal to zero
        for column = 1:D % corresponds to "z^(-1)" through "z^(-D)" in the transfer function
            k = column;
            y(column) = -(magnitude(m) * cos((-k * omega_ts) + phase(m)) * weight(m));
        end
        for column = 1:(N+1)   % corresponds to "z^(0)" through "z^(-N)" in the transfer function
            k = column - 1;
	        u(column) = cos(-k * omega_ts) * weight(m);    % yes, I know that cos(-x) = cos(x)
        end
        X(m,:) = [ y u ];   % [M x (D+N+1)]
    else    % denominator order equal to zero
        for column = 1:(N+1)   % corresponds to "z^(0)" through "z^(-N)" in the transfer function
            k = column - 1;
            u(column) = cos(-k * omega_ts) * weight(m);    % yes, I know that cos(-x) = cos(x)
        end
        X(m,:) = [ u ];   % [M x (N+1)]
    end
end

% Create Y vector
Y = magnitude .* cos(phase) .* weight;  % [M x 1]

% Choose your favorite matrix inversion method:
%	Compute Theta = [X]\[Y]; see Matlab documentation for "\" operator
Theta = X \ Y;    % [(D+N+1) x 1]
%	Compute Theta = [pinv(X)][Y]; see Matlab documentation for "pinv" operator
%Theta = pinv(X) * Y;    % [(D+N+1) x 1]
%	Compute Theta = [inv(X'X)][X'][Y]; traditional method -- not recommended
%Xt = X';    % [(D+N+1) x M]
%Theta = inv(Xt * X) * Xt * Y; % [(D+N+1) x 1] = [(D+N+1) x (D+N+1)] * [(D+N+1) x M] * [M x 1]

% Output numerator and denominator coefficients
B = Theta((D+1):(D+1+N))    % [(N+1) x 1]
if (D ~= 0)
    A = [ 1 ; Theta(1:D) ]  % [(D+1) x 1]
else
    A = [ 1 ]
end

% Plot the results (requires Signal Processing Toolbox)
inputmagnitude = 20.0 * log10(magnitude);
modelresponse = freqz(B,A,frequency,samplerate);
modelmagnitude = 20.0 * log10(abs(modelresponse));
modelphase = atan2(imag(modelresponse),real(modelresponse));
modelphase = modelphase + ((delta * TWOPI / samplerate) .* frequency);	% remove the effect of "delta", just for the plot
for m = 1:M
	while (modelphase(m) < -pi)
		modelphase(m) = modelphase(m) + TWOPI;
	end
	while (modelphase(m) > pi)
		modelphase(m) = modelphase(m) - TWOPI;
	end
end
modelphase = modelphase ./ DTR;
figure;
semilogx(frequency,modelmagnitude,'r');
hold on;
semilogx(frequency,inputmagnitude,'k');
grid;
hold off;
figure;
semilogx(frequency,modelphase,'r');
hold on;
semilogx(frequency,inputphase,'k');
grid;
hold off;

% Copyright  2006 Gregory Berchin, all rights reserved.  Copying permitted
% for personal, non-commercial use provided that this copyright statement
% remains intact.  Commercial use prohibited without expressed written
% consent from Gregory Berchin.
